#!/usr/bin/env python3

"""
\033[1;33mpicoslave\033[0m a CLI tool to control the \033[1;33mPicoSlave\033[0m I2C slave simulator.

PicoSlave USB devices have the vid/pid \033[0;33mC001:012C\033[0m and are automatically detected
by this tool if only one is present. The \033[1;39mscan\033[0m command lists devices showing an
index and the device serial number, by which a specific device can be selected
using \033[1;39m-i\033[0m or \033[1;39m-s\033[0m.

\033[1;33mPicoSlave\033[0m supports 2 I2C slaves to be each configured via the \033[1;39mconfig\033[0m command
for an I2C slave address to operate on a register space of configurable size
starting at address 0x00. Slave operation can be disabled by configuring a slave
for the I2C address 0x00. Slaves 0 or 1 are selected using the \033[1;39miface\033[0m argument
for all slave specific commands.

The register spaces of the slaves can be written into (\033[1;39mwrite\033[0m) and read from (\033[1;39mread\033[0m).
This can be done when slaves are enabled or disabled. \033[1;39mclear\033[0m resets the entire
configured register space to 0x00 or a given value.
"""  # pylint: disable=line-too-long  # noqa

import argparse
import logging
import sys
from typing import Any

import picoslave as ps
from picoslave.picoslave import PicoSlave, UsbPicoSlave
from picoslave.picoslave import bytestr
from picoslave.picoslave import DeviceNotFoundError


def config_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:

    function = PicoSlave.Protocol.SlaveFunction[args.function.upper()]

    if function == PicoSlave.Protocol.SlaveFunction.SLAVE:
        assert args.slave_addr is not None, "slave operation requires argument 'slave_addr'"
        slave_addr = args.slave_addr

        if args.eightbit:
            print(f"converting 8-bit address '{slave_addr:02X}h' to 7-bit address "
                  f'{slave_addr >> 1:02X}h')
            slave_addr >>= 1
        assert 1 <= slave_addr <= PicoSlave.Protocol.SlaveFunction.SLAVE.value, \
            f'I2C address out of range [1..{PicoSlave.Protocol.SlaveFunction.SLAVE:02X}h]'

        print(f'configuring I2C-{args.iface} for address {args.slave_addr:02X}h, '
              f'size {args.size} and word width {args.width}')
        picoslave.config_slave(args.iface, args.slave_addr, args.size, args.width)
        return

    if function == PicoSlave.Protocol.SlaveFunction.BLOCKER:
        assert args.signals, "'blocker' function requires '--signals' option (non-empty)."
        print(f'configuring I2C-{args.iface} as blocker ({"|".join(set(args.signals))})')
        picoslave.config_blocker(args.iface, scl='scl' in args.signals, sda='sda' in args.signals)
        return

    if function == PicoSlave.Protocol.SlaveFunction.RESET:
        args.slave_addr = PicoSlave.Protocol.SlaveFunction.RESET
        print(f'deactivating I2C-{args.iface}')
        picoslave.config_reset(args.iface)
        return


def read_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'reading from I2C-{args.iface}: {args.size} bytes from address {args.addr:02X}h')
    response_data = picoslave.read(args.iface, args.addr, args.size)
    print('response:')
    print(bytestr(response_data))


def write_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'writing to I2C-{args.iface}: {len(args.data)} '
          f'{"bytes" if len(args.data) > 1 else "byte"} to address {args.addr:02X}h:')
    print(bytestr(args.data))
    picoslave.write(args.iface, args.addr, args.data)


def clear_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'clearing memory of I2C-{args.iface} with value {args.value:02x}h')
    picoslave.clear(args.iface, args.value)


def stat_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'reading statistics from I2C-{args.iface}: {args.size} entries '
          f'from address {args.addr:02X}h')
    statistics = picoslave.statistics(args.iface, args.addr, args.size)
    print('  ADDR  READ  WRITE')
    for addr in statistics:
        if statistics[addr][0] or statistics[addr][1]:
            print(f'   {addr:02X}h  {statistics[addr][0]:4}   {statistics[addr][1]:4}')


def info_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    info = picoslave.info()
    print(info)


def reset_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print('resetting target')
    picoslave.reset()


def main() -> None:

    loglevels = ['debug', 'info', 'warning', 'error', 'critical']
    ifaces = [0, 1]

    def auto_int(x: str) -> int:
        return int(x, 0)

    def mem_addr8(x: str) -> int:
        addr = auto_int(x)
        if addr >= 2**8:
            raise argparse.ArgumentTypeError(f'address must be in range [0..{2**8-1}]')
        return addr

    def mem_addr16(x: str) -> int:
        addr = auto_int(x)
        if addr >= 2**16:
            raise argparse.ArgumentTypeError(f'address must be in range [0..{2**16-1}]')
        return addr

    def byte_str(x: str) -> bytes:
        try:
            return bytes.fromhex(x)
        except Exception:
            raise argparse.ArgumentTypeError(f'not a valid byte string: "{x}"')

    class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
        def add_usage(self, usage: Any, actions: Any, groups: Any, prefix: Any = None) -> None:
            if prefix is None:
                prefix = '\033[1;32mUsage: \033[0m'
            super().add_usage(usage, actions, groups, prefix)

    parser = argparse.ArgumentParser(description=__doc__,
                                     prog='\033[1;34mpicoslave\033[0m',
                                     formatter_class=Formatter)
    parser.add_argument(
        '-l', '--loglevel',
        metavar='L',
        type=str,
        default='Warning',
        choices=loglevels,
        help=f'log level: L={loglevels}\n'
    )
    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help="shortcut for '--loglevel debug'"
    )
    device_group = parser.add_mutually_exclusive_group()
    device_group.add_argument(
        '-i', '--index',
        type=int,
        metavar='X',
        help="Device index to select a specific PicoSlave device\n"
             "(see cmd='scan')"
    )
    device_group.add_argument(
        '-s', '--serial',
        type=str,
        metavar='S',
        help='Device serial number to select a specific PicoSlave\n'
             "device (see cmd='scan')"
    )
    parser.add_argument(
        '-V', '--version',
        action='store_true',
        help='show the program version'
    )

    # subparsers
    sub = parser.add_subparsers(
        title='\033[1;34mCMD\033[0m',
        dest='command',
        metavar='\033[1;34mCMD\033[0m',
        description='Command to execute on PicoSlave. See CMD -h for help.',
    )
    iface_parser = argparse.ArgumentParser(add_help=False, formatter_class=Formatter)
    iface_parser.add_argument(
        'iface',
        type=int,
        metavar='iface',
        choices=ifaces,
        help=f'I2C interface number ({ifaces})',
    )

    # config
    sub_parser = sub.add_parser(
        'config',
        aliases=['c'],
        parents=[iface_parser],
        help='configure an I2C slave',
        formatter_class=Formatter,
    )
    sub_parser.add_argument(
        '-f', '--function',
        metavar='FUNC',
        type=str,
        default='slave',
        choices=[e.name.lower() for e in PicoSlave.Protocol.SlaveFunction],
        help='Function to configure the given interface.\n'
             "    slave:   (default) operate as I2C slave\n"
             "             requires 1 <= 'slave_addr' <= 127\n"
             "    blocker: block SCL, SDA or both to a LOW signal level\n"
             "             requires '--signals' option\n"
             "    reset:   reset configuration for no operation\n",
    )
    sub_parser.add_argument(
        'slave_addr',
        type=auto_int,
        nargs='?',
        help='Memory or 7-bit slave address (use -8 for 8-bit addresses)',
    )
    sub_parser.add_argument(
        '-8',
        dest='eightbit',
        action='store_true',
        help='Set to interpret I2C addresses as 8-bit (default is 7-bit)\n',
    )
    sub_parser.add_argument(
        'size',
        type=int,
        nargs='?',
        default=256,
        help='size of the configured I2C memory, at which it will overflow\n'
             '(max 256). Note that the I2C memory size is given in words\n'
             '(of size [width]), not bytes.',
    )
    sub_parser.add_argument(
        'width',
        type=int,
        nargs='?',
        default=1,
        choices=[1, 2, 4],
        metavar='width',
        help='width of data words in I2C memory (1, 2 or 4)',
    )
    sub_parser.add_argument(
        '--signals',
        type=str,
        nargs='*',
        choices=['sda', 'scl'],
        metavar='S',
        help='todo',
    )
    sub_parser.set_defaults(func=config_cmd)

    # read
    sub_parser = sub.add_parser(
        'read',
        aliases=['r'],
        parents=[iface_parser],
        help='read from the slaves I2C memory',
        formatter_class=Formatter,
    )
    sub_parser.add_argument(
        'addr',
        type=mem_addr16,
        help='address to start reading from',
    )
    sub_parser.add_argument(
        'size',
        type=int,
        nargs='?',
        default=1,
        help='number of words to read from [addr] (max 256 * [width])',
    )
    sub_parser.set_defaults(func=read_cmd)

    # write
    sub_parser = sub.add_parser(
        'write',
        aliases=['w'],
        parents=[iface_parser],
        help='write data to the slaves I2C memory',
        formatter_class=Formatter,
    )
    sub_parser.add_argument(
        'addr',
        type=mem_addr16,
        help='address to start writing into',
    )
    sub_parser.add_argument(
        'data',
        type=byte_str,
        help='data to write to the I2C memory at [addr].\n'
             'Data format is a little endian hex string without spaces,\n'
             'e.g.: 03ab7fE0 (4 bytes). Note that data length must be a\n'
             'multiple of the configured word [width]. (see "config -h")',
    )
    sub_parser.set_defaults(func=write_cmd)

    # clear
    sub_parser = sub.add_parser(
        'clear',
        aliases=['C'],
        parents=[iface_parser],
        help='clear the I2C memory to 0 or an initial value',
        formatter_class=Formatter,
    )
    sub_parser.add_argument(
        'value',
        type=auto_int,
        nargs='?',
        default=0x00,
        help='initialize with this value',
    )
    sub_parser.set_defaults(func=clear_cmd)

    # stat
    sub_parser = sub.add_parser(
        'stat',
        aliases=['s'],
        parents=[iface_parser],
        help='retrieve read/write statistics from I2C memory',
        formatter_class=Formatter,
    )
    sub_parser.add_argument(
        'addr',
        type=mem_addr8,
        nargs='?',
        default=0x00,
        help='start memory address for reading statistics',
    )
    sub_parser.add_argument(
        'size',
        type=int,
        nargs='?',
        default=256,
        help='number of addresses/words to get r/w statistics for (max 256)',
    )
    sub_parser.set_defaults(func=stat_cmd)

    # scan
    sub_parser = sub.add_parser(
        'scan',
        aliases=['S'],
        help='scan for PicoSlave devices on USB',
        formatter_class=Formatter,
    )
    sub_parser.set_defaults(func='scan_cmd')

    # info
    sub_parser = sub.add_parser(
        'info',
        aliases=['I'],
        help='print info for the selected device',
        formatter_class=Formatter,
    )
    sub_parser.set_defaults(func=info_cmd)

    # reset
    sub_parser = sub.add_parser(
        'reset',
        aliases=['R'],
        help='reset the PicoSlave device',
        formatter_class=Formatter,
    )
    sub_parser.set_defaults(func=reset_cmd)

    args = parser.parse_args()

    if args.version:
        print(f'PicoSlave {ps.__version__}')
        exit(0)

    if args.debug:
        args.loglevel = 'debug'

    log_format = '%(asctime)s %(name)-10s %(levelname)-8s %(message)s'
    logging.basicConfig(level=logging.getLevelName(args.loglevel.upper()), format=log_format)

    try:
        import coloredlogs  # type: ignore
        coloredlogs.install(level=logging.getLevelName(args.loglevel.upper()), fmt=log_format)
    except ModuleNotFoundError:
        pass

    if 'func' not in args:
        parser.print_help()
        exit(1)

    if args.func == 'scan_cmd':
        devices = UsbPicoSlave.scan()
        sys.exit(1 if not devices else 0)

    assert args.index is None or args.index > 0, 'Device index must not be 0'
    args.index = args.index - 1 if args.index is not None else None

    picoslave = PicoSlave(index=args.index, serial=args.serial)

    try:
        args.func(args, picoslave)
    except (AssertionError, DeviceNotFoundError) as e:
        if str(e):
            print(f'\033[1;31mError:\033[0m {e}')
            sys.exit(1)
        raise


if __name__ == '__main__':
    main()



from enum import IntEnum
import logging
import struct
from typing import List, Dict, Tuple, Any, Optional, cast

import usb.core  # type: ignore
import usb.util  # type: ignore
import usb.legacy  # type: ignore


log = logging.getLogger('picoslave')


def bytestr(b: bytes) -> str:
    assert b is not None
    return ' '.join(f'{x:02X}' for x in b)


class DeviceNotFoundError(Exception):
    """error when a specified device was not found"""


class UsbError(Exception):
    """error when communication with the device was not possible"""


class ProtocolError(Exception):
    """when unexpected respnses are received from the device"""


class UsbPicoSlave:

    ID_VENDOR = 0xC001
    ID_PRODUCT = 0x012C
    OUT_EP_ADDR = 0x04
    IN_EP_ADDR = 0x85

    def __init__(self, device: usb.core.Device) -> None:
        assert device is not None

        cfg = device.get_active_configuration()
        if cfg is None:
            raise UsbError('no active configuration on device')

        intf = usb.util.find_descriptor(cfg, bInterfaceClass=usb.legacy.CLASS_VENDOR_SPEC)
        if intf is None:
            raise UsbError("couldn't find PicoSlave interface on device")

        outep = usb.util.find_descriptor(intf, bEndpointAddress=self.OUT_EP_ADDR)
        if outep is None:
            raise UsbError(f"OUT endpoint (0x{self.OUT_EP_ADDR:02X}) not found on device")

        inep = usb.util.find_descriptor(intf, bEndpointAddress=self.IN_EP_ADDR)
        if inep is None:
            raise UsbError(f"IN endpoint (0x{self.IN_EP_ADDR:02X}) not found on device")

        self._device = device
        self._inep = inep
        self._outep = outep

    def close(self) -> None:
        try:
            self._device.reset()
        except usb.core.USBError as exc:
            raise UsbError(f'Failed to close the device ({exc})') from exc

    def write(self, data: bytes) -> None:
        """raises: UsbError"""
        try:
            wire_hdr = struct.pack('<I', len(data))
            self._outep.write(wire_hdr + data)
        except usb.core.USBError as exc:
            raise UsbError(f'Failed to write to device ({exc})') from exc

    def read(self) -> bytes:
        """raises: UsbError"""
        try:
            packet = self._inep.read(8192)
            len_hdr = struct.unpack('<I', packet[:4])[0]
            data = packet[4:]
            log.debug(f'[rx]: len_hdr: {len_hdr}, len(data): {len(data)}')
            if len_hdr != len(data):
                raise ProtocolError(
                    f'received wrong number of bytes. header: {len_hdr}, data: {len(data)}'
                )
            return bytes(list(data))
        except usb.core.USBError as exc:
            raise UsbError(f'No response from device ({exc})') from exc

    @classmethod
    def _scan(cls) -> List[usb.core.Device]:
        devs = usb.core.find(idVendor=cls.ID_VENDOR, idProduct=cls.ID_PRODUCT, find_all=True)
        devices: List[usb.core.Device] = []
        for d in devs or []:
            try:
                # try to access the serial number.
                # this gives an error when the device is in a weird state
                log.debug(f'found device with serial number: {d.serial_number}')
            except Exception as exc:
                log.warning(f"couldn't read from device: {exc}")
                continue
            devices.append(d)
        return sorted(devices, key=lambda d: cast(int, d.serial_number))

    @classmethod
    def scan(cls) -> List[usb.core.Device]:
        devices = cls._scan()
        vid_pid = f'{cls.ID_VENDOR:04x}:{cls.ID_PRODUCT:04x}'
        id = 1
        for d in devices:
            usb_path = f'{d.bus}.{".".join(str(p) for p in d.port_numbers)}:{d.address}'
            print(f'{id:>2}  {d.serial_number:<16}  {usb_path:>12}-{vid_pid:<9}'
                  f'  {d.manufacturer} {d.product} ')
            id += 1
        if not devices:
            print(f'no PicoSlave device found with {vid_pid}')
        return devices

    @classmethod
    def find(cls, index: Optional[int] = None, serial: Optional[str] = None) -> 'UsbPicoSlave':
        assert index is None or serial is None, "'index' and 'serial' must not be both set"

        selected_device: usb.core.Device = None
        devices = cls._scan()

        if index is not None:
            if index >= len(devices):
                raise DeviceNotFoundError(f'No device with index "{index}"')
            selected_device = devices[index]
        elif serial is not None:
            matching = [dev for dev in devices if dev.serial_number == serial]
            if not matching:
                raise DeviceNotFoundError(f'No device found with serial number "{serial}"')
            if len(matching) > 1:
                raise DeviceNotFoundError(f'Multiple devices found with serial number "{serial}" '
                                          f'({len(matching)}). Select with "index" filter')
            selected_device = matching[0]
        else:
            if len(devices) == 0:
                raise DeviceNotFoundError('No device found')
            if len(devices) > 1:
                raise DeviceNotFoundError(f'Multiple devices found ({len(devices)}). '
                                          'Select with "index" or "serial" filter.')
            selected_device = devices[0]
        return UsbPicoSlave(selected_device)


class PicoSlave:
    class Protocol:

        MAX_RW_SIZE = 32

        class Command(IntEnum):
            CONFIG = 0xA0
            READ = 0xA1
            WRITE = 0xA2
            CLEAR = 0xA3
            STAT = 0xA4
            INFO = 0xB0
            RESET = 0xBF

        class ResponseCode(IntEnum):
            OK = 0
            CRC_ERROR = 1
            INVALID_PACKET = 2
            INVALID_REQUEST = 3
            INVALID_INTERFACE = 4
            INVALID_ADDRESS = 5
            INVALID_SIZE = 6
            INVALID_WIDTH = 7
            MEMORY_ERROR = 8
            OPERATION_FAILED = 9

        class SlaveFunction(IntEnum):
            RESET = 0,
            SLAVE = 0x7F,
            BLOCKER = 0xF1,

        @staticmethod
        def crc16(data: bytes, poly: int = 0x8408) -> int:
            """CRC-16-CCITT Algorithm
            source: https://gist.github.com/oysstu/68072c44c02879a2abf94ef350d1c7c6"""
            crc = 0xFFFF
            for b in data:
                cur_byte = 0xFF & b
                for _ in range(0, 8):
                    if (crc & 0x0001) ^ (cur_byte & 0x0001):
                        crc = (crc >> 1) ^ poly
                    else:
                        crc >>= 1
                    cur_byte >>= 1
            crc = (~crc & 0xFFFF)
            crc = (crc << 8) | ((crc >> 8) & 0xFF)
            return crc & 0xFFFF

        @classmethod
        def create_packet(cls, cmd: Command, iface: int, addr: int, size: int = 0,
                          data: Optional[bytes] = None) -> bytes:
            assert size < 2**16, 'size must be in range [0..2**16]'
            assert addr < 2**16, 'addr must be in range [0..2**16]'

            if cmd == cls.Command.CONFIG:
                assert size == len(data or []), \
                    f"size ({size}) doesn't match data size ({len(data or [])})"
            elif cmd == cls.Command.READ or cmd == cls.Command.STAT:
                assert size > 0, 'READ/STAT packet needs size > 0'
                assert data is None, 'READ/STAT must not have data'
            elif cmd == cls.Command.WRITE:
                assert data is not None, 'WRITE packet needs data'
                size = len(data)
            elif cmd == cls.Command.CLEAR:
                assert size == 0, 'CLEAR size must be 0'
                assert data is None, 'CLEAR must not have data'
            elif cmd == cls.Command.RESET:
                assert size == 0, 'RESET size must be 0'
                assert data is None, 'RESET must not have data'
            elif cmd == cls.Command.INFO:
                assert size == 0, 'INFO size must be 0'
                assert data is None, 'INFO must not have data'
            else:  # pragma: no cover
                assert False, f'Unsupported CMD: {cmd}'
            header = bytes([int(cmd), iface]) + struct.pack('<H', addr) + struct.pack('<H', size)
            packet = header + (data or bytes([]))
            crc = struct.pack('<H', cls.crc16(packet))
            log.debug(f'packet+crc: {bytestr(packet + crc)}')
            return packet + crc

        @classmethod
        def verify(cls, packet: bytes) -> bytes:
            # check that the packet is long enough to make any sense
            if len(packet) < 5:
                raise ProtocolError(f'invalid received packet size: {len(packet)}')

            # check CRC before looking into the packet
            crc_calc = cls.crc16(packet[:-2])
            crc_rx = struct.unpack('<H', packet[-2:])[0]
            if crc_calc != crc_rx:
                raise ProtocolError(f'CRC mismatch: actual: {crc_rx:04X}, expected: {crc_calc:04X}')
            log.debug('rx packet CRC check successful')
            packet_data = packet[:-2]

            # check the result code
            try:
                code = packet_data[0]
                response_code = cls.ResponseCode(code)
                if response_code != cls.ResponseCode.OK:
                    raise ProtocolError(f'received error code {code} ({response_code.name})')
            except ValueError as exc:
                raise ProtocolError(f'received unknown response code: {code}') from exc

            # check the packet data size (-3 is the header size)
            size = struct.unpack('<H', packet_data[1:3])[0]
            if size != len(packet_data) - 3:
                raise ProtocolError(f'received packet size mismatch: size={size}, '
                                    f'data={len(packet_data) - 3}')
            return packet_data[3:]

    def __init__(self, index: Optional[int] = None, serial: Optional[str] = None) -> None:
        assert index is None or serial is None, "'index' and 'serial' must not be both set"
        self._usb = UsbPicoSlave.find(index=index, serial=serial)

    def close(self) -> None:
        self._usb.close()

    def _txrx(self, packet: bytes) -> bytes:
        """raises: UsbError"""

        log.debug(f'tx-packet: {bytestr(packet)}')
        self._usb.write(packet)

        response = self._usb.read()
        log.debug(f'rx-packet: {bytestr(response)}')

        rx_data = self.Protocol.verify(response)
        log.debug(f'rx-data: {bytestr(rx_data)}')
        return rx_data

    def config(self, iface: int, slave_address: int, mem_size: int = 256,
               mem_width: int = 1) -> None:
        """Legacy"""
        if slave_address == PicoSlave.Protocol.SlaveFunction.RESET:
            self.config_reset(iface)
        elif slave_address <= PicoSlave.Protocol.SlaveFunction.SLAVE:
            self.config_slave(iface, slave_address, mem_size, mem_width)
        elif slave_address == PicoSlave.Protocol.SlaveFunction.BLOCKER:
            self.config_blocker(iface, scl=True, sda=True)
        else:
            # pass through using this function as a raw configuration interface
            self._config(iface, slave_address, data=struct.pack('<HH', mem_size, mem_width))

    def config_slave(self, iface: int, slave_address: int, mem_size: int = 256,
                     mem_width: int = 1) -> None:
        """Initialize or re-initialize the slave for a given I2C address.
        Address 0 deactivates the slave. """
        assert 1 <= slave_address <= PicoSlave.Protocol.SlaveFunction.SLAVE, \
            f"'slave_address' must be in range [1..{PicoSlave.Protocol.SlaveFunction.SLAVE}]"
        assert mem_size > 0
        assert mem_width in [1, 2, 4]
        log.info(f'config: iface={iface}, slave_address={slave_address:02X}h, '
                 f'mem_size={mem_size}, mem_width={mem_width}')
        self._config(
            iface=iface,
            addr=slave_address,
            data=struct.pack('<HH', mem_size, mem_width),
        )

    def config_blocker(self, iface: int, scl: bool, sda: bool) -> None:
        self._config(
            iface=iface,
            addr=PicoSlave.Protocol.SlaveFunction.BLOCKER,
            data=bytes([int(scl), int(sda)]),
        )

    def config_reset(self, iface: int) -> None:
        self._config(
            iface=iface,
            addr=PicoSlave.Protocol.SlaveFunction.RESET,
            data=bytes(0)
        )

    def _config(self, iface: int, addr: int, data: bytes) -> None:
        packet = self.Protocol.create_packet(
            self.Protocol.Command.CONFIG, iface, addr, size=len(data), data=data
        )
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'CONFIG received an unexpected response: {bytestr(rsp)}')

    def read(self, iface: int, mem_addr: int, size: int) -> bytes:
        log.info(f'read: iface={iface}, mem_addr={mem_addr:02X}h, size={size}')
        packet = self.Protocol.create_packet(self.Protocol.Command.READ, iface, mem_addr, size)
        rsp = self._txrx(packet)
        if not rsp:
            raise ProtocolError('received no data')
        return rsp

    def write(self, iface: int, mem_addr: int, data: bytes) -> None:
        log.info(f'write: iface={iface}, mem_addr={mem_addr:02X}h, data_size={len(data)}, '
                 f'data="{bytestr(data)}"')
        packet = self.Protocol.create_packet(self.Protocol.Command.WRITE, iface, mem_addr,
                                             len(data), data)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'WRITE received an unexpected response: {bytestr(rsp)}')

    def clear(self, iface: int, reset_value: int = 0) -> None:
        log.info(f'clear: iface={iface}, reset_value={reset_value}')
        packet = self.Protocol.create_packet(self.Protocol.Command.CLEAR, iface, reset_value)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'CLEAR received an unexpected response: {bytestr(rsp)}')

    def statistics(self, iface: int, mem_addr: int, size: int) -> Dict[int, Tuple[int, int]]:
        log.info(f'stat: iface={iface}, mem_addr={mem_addr:02X}h, size={size}')
        packet = self.Protocol.create_packet(self.Protocol.Command.STAT, iface, mem_addr, size)
        rsp = self._txrx(packet)
        if not rsp:
            raise ProtocolError('received no data')
        if len(rsp) % 8 != 0:
            raise ProtocolError(f'received data size is no multiple of {8} (size={len(rsp)})')

        statistics = {}
        stat_addr = mem_addr
        for i in range(0, len(rsp), 8):
            read_cnt = struct.unpack('<I', rsp[i:i + 4])[0]
            write_cnt = struct.unpack('<I', rsp[i + 4:i + 8])[0]
            statistics[stat_addr] = (read_cnt, write_cnt)
            stat_addr += 1
        return statistics

    def info(self) -> Dict[str, Any]:
        info_segments = 4
        log.info('reading device info')
        packet = self.Protocol.create_packet(self.Protocol.Command.INFO, 0, 0)
        rsp = self._txrx(packet)

        if not rsp:
            raise ProtocolError('received no data')

        try:
            info = rsp.decode('utf-8')
        except UnicodeDecodeError as exc:
            raise ProtocolError(f'failed to decode INFO response: {rsp!r}') from exc

        infos = info.split(';')
        if len(infos) != info_segments:
            raise ProtocolError('received unexpected number of INFO segments '
                                f'(actual: {len(infos)}, expected: {info_segments})')

        return dict(
            serial=infos[0],
            firmware=infos[1],
            protocol=infos[2],
            interfaces=infos[3],
        )

    def reset(self) -> None:
        """Reset the device by sending the `RESET` command to it.

        It is expected that the device resets and does not respond, otherwise the reset is
        considered to have failed.

        Raises:
            ProtocolError: When the device unexpectedly responds.
        """
        log.info('reset device')
        packet = self.Protocol.create_packet(self.Protocol.Command.RESET, 0, 0)
        try:
            rsp = self._txrx(packet)
            if rsp:
                raise ProtocolError(f'RESET received an unexpected response: {bytestr(rsp)}')
        except UsbError as exc:
            log.warning(f'error when sending reset packet: {str(exc)}')
        finally:
            self.close()
